"""
This implements the paper 'Distance Transforms of Sampled Functions'
by Felzenszwalb and Huttenlocher.

Distance fields are useful in a variety of applications, including
image segmentation, antialiasing algorithms, and texture synthesis.
"""

from numba import jit
import numba
import numpy as np
from . import io

INF = 1e20

def generate_sdf(image: np.ndarray, wrapx=False, wrapy=False):
    """Create a signed distance field from a boolean field."""
    a = generate_udf(image, wrapx, wrapy)
    b = generate_udf(image == 0.0, wrapx, wrapy)
    return a - b

def generate_udf(image: np.ndarray, wrapx=False, wrapy=False):
    """Create an unsigned distance field from a boolean field."""
    assert image.dtype == 'bool', 'Pixel values must be boolean'
    assert len(image.shape) == 3, 'Shape is not rows x cols x channels'
    assert image.shape[2] == 1, 'Image must be grayscale'
    return _generate_edt(image, wrapx, wrapy)

def generate_gdf(image: np.ndarray, wrapx=False, wrapy=False):
    "Create an generalized squared distance field from a scalar field."
    assert image.dtype == 'float64', 'Pixel values must be real'
    assert len(image.shape) == 3, 'Shape is not rows x cols x channels'
    assert image.shape[2] == 1, 'Image must be grayscale'
    return _generate_gdt(image, wrapx, wrapy)

def _generate_gdt(image, wrapx, wrapy):
    image = io.unshape(image)
    height, width = image.shape
    capacity = max(width, height)
    if wrapx or wrapy: capacity *= 2
    d = np.zeros([capacity])
    z = np.zeros([capacity + 1])
    v = np.zeros([capacity], dtype='i4')
    result = image.copy()
    _generate_udf(width, height, d, z, v, result, wrapx, wrapy)
    return io.reshape(result)

def _generate_edt(image, wrapx, wrapy):
    image = io.unshape(image)
    height, width = image.shape
    capacity = max(width, height)
    if wrapx or wrapy: capacity *= 2
    d = np.zeros([capacity])
    z = np.zeros([capacity + 1])
    v = np.zeros([capacity], dtype='i4')
    result = np.where(image, 0.0, INF)
    _generate_udf(width, height, d, z, v, result, wrapx, wrapy)
    return np.sqrt(io.reshape(result))

@jit(nopython=True, fastmath=True)
def _generate_udf(width, height, d, z, v, result, wrapx, wrapy):
    # Compute 1D distance fields for columns, then for rows.
    for x in range(width):
        f = result[:,x]
        edt(f, d, z, v, height, wrapy)
        result[:,x] = d[:height]
    for y in range(height):
        f = result[y,:]
        edt(f, d, z, v, width, wrapx)
        result[y,:] = d[:width]

@jit(nopython=True, fastmath=True)
def edt(f, d, z, v, n, wrap):
    # Find the lower envelope of a sequence of parabolas.
    #   f...source data (returns the Y of the parabola rooted at X)
    #   d...destination data (final distance values are written here)
    #   z...temporary used to store X coords of parabola intersections
    #   v...temporary used to store X coords of parabola roots
    #   n...number of pixels in "f" to process

    # Always add the first pixel to the enveloping set since it is
    # obviously lower than all parabolas processed so far.
    k: int = 0
    v[0] = 0
    z[0] = -INF
    z[1] = +INF
    upper = 2*n if wrap else n

    for q in range(1, upper):

        # If the new parabola is lower than the right-most parabola in
        # the envelope, remove it from the envelope. To make this
        # determination, find the X coordinate of the intersection (s)
        # between the parabolas rooted at (q,f[q]) and (p,f[p]).
        p = v[k]
        s = ((f[q%n] + q*q) - (f[p%n] + p*p)) / (2.0*q - 2.0*p)
        while s <= z[k]:
            k = k - 1
            p = v[k]
            s = ((f[q%n] + q*q) - (f[p%n] + p*p)) / (2.0*q - 2.0*p)

        # Add the new parabola to the envelope.
        k = k + 1
        v[k] = q
        z[k] = s
        z[k + 1] = +INF

    # Go back through the parabolas in the envelope and evaluate them
    # in order to populate the distance values at each X coordinate.
    k = 0
    lower = n/2 if wrap else 0
    upper = 3*n/2 if wrap else n
    for q in range(lower, upper):
        while z[k + 1] < float(q):
            k = k + 1
        vmod = v[k] % n
        qmod = q % n
        dx = q - v[k]
        d[qmod] = dx * dx + f[vmod]



"""Define resize, blur, and related constants."""

from . import io
from collections import namedtuple
from numba import guvectorize
import math
import numpy as np

RowOps = namedtuple('RowOps', 'tindices sindices fweights'.split())

GAUSSIAN_SCALE = 1.0 / np.sqrt(0.5 * np.pi)

def hermite(x):
    x = np.clip(x, 0, 1)
    return 2 * x * x * x - 3 * x * x + 1

def triangle(x):
    x = np.clip(x, 0, 1)
    return 1.0 - x

def gaussian(x):
    x = np.clip(x, 0, 2)
    return np.exp(-2 * x * x) * GAUSSIAN_SCALE

def nearest(x):
    return np.less_equal(x, 0.5) * 1.0

def sinc(x):
    if x <= 0.00001: return 1.0
    return np.sin(np.pi * x) / (np.pi * x)

def lanczos(x):
    x = np.clip(x, 0, 1)
    return sinc(x) * sinc(x)

def mitchell(x):
    B = 1.0 / 3.0
    C = 1.0 / 3.0
    P0 = (6 - 2*B) / 6.0
    P1 = 0
    P2 = (-18 +12*B + 6*C) / 6.0
    P3 = (12 - 9*B - 6*C) / 6.0
    Q0 = (8*B +24*C) / 6.0
    Q1 = (-12*B -48*C) / 6.0
    Q2 = (6*B +30*C) / 6.0
    Q3 = (-1*B - 6*C) / 6.0
    if x >= 2.0: return 0.0
    if x >= 1.0: return Q0 + Q1*x + Q2*x*x + Q3*x*x*x
    return P0 + P1*x + P2*x*x + P3*x*x*x

class Filter:
    def __init__(self, fn, radius):
        self.radius = radius
        self.function = fn

HERMITE  = Filter(hermite, 1)
TRIANGLE = Filter(triangle, 1)
GAUSSIAN = Filter(gaussian, 2)
NEAREST  = Filter(nearest, 0)
LANCZOS  = Filter(lanczos, 1)
MITCHELL = Filter(mitchell, 2)

def resize(source, width=None, height=None, filter=None, radius=1,
           wrapx=False, wrapy=False):
    """Create a new numpy image with the desired size.

    Either width or height can be null, in which case its value
    is inferred from the aspect ratio of the source image.

    Filter can be HERMITE, TRIANGLE, GAUSSIAN, NEAREST, LANCZOS, or
    MITCHELL.
    """
    assert len(source.shape) == 3, 'Shape is not rows x cols x channels'
    assert width != None or height != None,  'Missing target size'
    aspect = source.shape[1] / source.shape[0]
    if width == None: width = height * aspect
    if height == None: height = width / aspect
    magnifying = width > source.shape[1]
    if filter == None: filter = MITCHELL if magnifying else LANCZOS
    return resample(source, width, height, filter, radius, wrapx, wrapy)

def resample(source, width, height, filter, radius, wrapx, wrapy):
    nchans = source.shape[2]
    def fn(t): return filter.function(t / radius)
    scaled_filter = Filter(fn, radius * filter.radius)
    srows, scols = source.shape[0], source.shape[1]
    trows, tcols = int(height), int(width)
    vresult = np.zeros([srows, tcols, nchans])
    rowops = create_ops(tcols, scols, scaled_filter, wrapx)
    convolve(vresult, source, rowops)
    vresult = transpose(vresult)
    hresult = np.zeros([tcols, trows, nchans])
    rowops = create_ops(trows, srows, scaled_filter, wrapy)
    convolve(hresult, vresult, rowops)
    return transpose(hresult)

def blur(image, filter=GAUSSIAN, radius=4, wrapx=False, wrapy=False):
    """Resample an image and produce a new image with the same size.
    
    For a list of available filters, see <a href="#resize">resize</a>.
    """
    width, height = image.shape[1], image.shape[0]
    return resize(image, width, height, filter, radius, wrapx, wrapy)

def transpose(source: np.ndarray):
    return np.swapaxes(source, 0, 1)

def create_ops(ntarget, nsource, filter: Filter, wrap) -> RowOps:
    # Generate a sequence of operations to perform a 1D convolution
    # where each operation is represented by 3-tuple of: target index,
    # source index, weight.
    tindices, sindices, fweights = [], [], []
    dtarget = 1.0 / ntarget
    dsource = 1.0 / nsource
    minifying = ntarget < nsource
    fextent = dtarget if minifying else dsource
    fdomain = float(ntarget if minifying else nsource)
    x = dtarget / 2
    for tindex in range(ntarget):
        minx = x - filter.radius * fextent
        maxx = x + filter.radius * fextent
        minsi = int(minx * float(nsource))
        maxsi = int(math.ceil(maxx * float(nsource)))
        localops = []
        weightsum = 0.0
        for sindex in range(minsi, maxsi+1):
            wrapped = sindex
            if sindex < 0 or sindex >= nsource:
                if wrap:
                    wrapped = sindex % nsource
                else:
                    continue
            sx = (0.5 + sindex) * dsource
            t = fdomain * abs(sx - x)
            weight = filter.function(t)
            if weight != 0:
                localops.append((tindex, wrapped, weight))
                weightsum += weight
        if weightsum > 0.0:
            for op in localops:
                tindices.append(op[0])
                sindices.append(op[1])
                fweights.append(op[2] / weightsum)
        x += dtarget
    return RowOps(tindices, sindices, fweights)

SIG0 = "void(f8[:,:,:], f8[:,:,:], i4[:], i4[:], f8[:])"
SIG1 = "(r0,c0,d),(r0,c1,d),(i),(i),(i)"
@guvectorize([SIG0], SIG1, target='parallel')
def jit_convolve(target, source, tinds, sinds, weights):
    nrows, nchan, nops = target.shape[0], target.shape[2], len(tinds)
    for c in range(nchan):
        for row in range(nrows):
            for op in range(nops):
                tind, sind, weight = tinds[op], sinds[op], weights[op]
                target[row][tind][c] += source[row][sind][c] * weight

def convolve(target, source, rowops: RowOps):
    # Perform highly generalized 1D convolution. This is almost
    # equivalent to:
    #
    # for row in range(len(target)):
    #     target[row][tindices] += source[row][sindices] * fweights
    #
    # ...but with the crucial feature of allowing the same index to
    # appear multiple times in tindices.
    #
    # Note that standard numpy convolution assumes a stationary kernel,
    # whereas this function could possibly be used to apply a varying
    # kernel.
    tindices, sindices, fweights = rowops
    assert len(tindices) == len(sindices) == len(fweights)
    assert len(target) == len(source)
    jit_convolve(target, source,
                 np.int32(tindices), np.int32(sindices),
                 np.double(fweights))



"""Define export, load, show, reshape and unshape."""

import imageio
import numpy as np
import os
import platform
import sys
import tempfile

LINEAR = 0
SRGB = 1
GAMMA = 2

def sRGB_to_linear(s):
   a = 0.055
   return np.where(s <= 0.04045, s / 12.92, ((s+a) / (1+a)) ** 2.4)

def linear_to_sRGB(s):
   a = 0.055
   return np.where(s <= 0.0031308, 12.92 * s, (1+a) * s**(1/2.4) - a)

def gamma_to_linear(s):
   return s ** 2.2

def linear_to_gamma(s):
   return s ** (1/2.2)

def linearize(image, target_space=SRGB):
    """Transform colors from perceptually linear to physically linear.
    
    This is automatically performed when using <a href="#load">load</a>
    on a PNG or JPEG. See also <a href="#delinearize">delinearize</a>.
    """
    if target_space == SRGB:
        return sRGB_to_linear(image)
    return gamma_to_linear(image)

def delinearize(image, source_space=SRGB):
    """Transform colors from physically linear to perceptually linear.
    
    This is automatically performed when using <a href="#export">export</a>
    to a PNG or JPEG. See also <a href="#linearize">linearize</a>.
    """
    if source_space == SRGB:
        return linear_to_sRGB(image)
    return linear_to_gamma(image)

def show(image, delinearize=True):
    """Display an image in a platform-specific way."""
    if isinstance(image, np.ndarray):
        show_array(image, delinearize)
    elif isinstance(image, str):
        show_filename(image)
    else:
        raise ValueError('Unsupported type')

def reshape(image):
    """Add a trailing dimension to single-channel 2D images.

    See also <a href="#unshape">unshape</a>.
    """
    if len(image.shape) == 2:
        image = np.reshape(image, image.shape + (1,))
    return image

def unshape(image):
    """Remove the trailing dimension from single-channel 3D images.
    
    See also <a href="#reshape">reshape</a>.
    """
    if len(image.shape) == 3 and image.shape[2] == 1:
        return np.reshape(image, image.shape[:2])
    return image

def _load(filename: str, linear: bool):
    if filename.endswith('.png'):
        img = imageio.imread(filename, 'PNG-PIL', pilmode='RGBA')
        img = np.clip(np.float64(img) / 255, 0, None)    
    elif filename.endswith('.jpg') or filename.endswith('.jpeg'):
        img = imageio.imread(filename)
        img = np.clip(np.float64(img) / 255, 0, None)
    elif filename.endswith('.exr'):
        imageio.plugins.freeimage.download()
        img = np.float64(imageio.imread(filename))
    return linearize(img) if not linear else img

def load(filename: str, linearize=True) -> np.ndarray:
    """Create a numpy array from the given PNG, JPEG, or EXR image file.

    Regardless of the pixel format on disk, PNG / JPEG images are always
    divided by 255, and PNG images are extended to 4 color channels.

    See also <a href="#reshape">reshape</a> and
    <a href="#linearize">linearize</a>  (which this calls).
    """
    assert filename.endswith('.png') or filename.endswith('.jpeg') or \
            filename.endswith('.jpg') or filename.endswith('.exr')
    return reshape(np.float64(_load(filename, not linearize)))

def _export(image: np.ndarray, filename: str, linear):
    image_format = None
    if linear:
        image = delinearize(np.clip(image, 0, None))
    if filename.endswith('.exr'):
        imageio.plugins.freeimage.download()
        image_format = 'EXR-FI'
        image = np.float32(image)
    else:
        image = np.uint8(np.clip(image * 255, 0, 255))
    imageio.imwrite(filename, unshape(image), image_format)

def export(image: np.ndarray, filename: str, delinearize=True):
    """Export a numpy array to a PNG, JPEG, or EXR image file.

    This function automatically multiplies PNG / JPEG images by 255.

    See also <a href="#unshape">unshape</a> and
    <a href="#delinearize">delinearize</a> (which this calls).
    """
    assert filename.endswith('.png') or filename.endswith('.jpeg') or \
            filename.endswith('.jpg') or filename.endswith('.exr')
    _export(image, filename, delinearize)

def show_array(image: np.ndarray, delinearize):
    with tempfile.NamedTemporaryFile() as fp:
        filename = fp.name + '.png'
        export(image, filename, delinearize)
        show_filename(filename)

def show_filename(image: str):
    if 0 == os.system('which imgcat >/dev/null 2>&1'):
        print("\n")
        os.system('imgcat ' + image)
    elif 'pytest' in sys.modules:
        return
    elif platform.system() == 'Darwin':
        os.system('open ' + image)
    elif platform.system() == 'Linux' and \
            os.environ.get('DESKTOP_SESSION'):
        os.system('xdg-open ' + image)
    else:
        print('Generated ' + image)

def ensure_alpha(src: np.ndarray) -> np.ndarray:
    """If the incoming image is 3-channel, adds a 4th channel."""
    assert len(src.shape) == 3
    if src.shape[2] != 3:
        return src
    alpha = np.ones(src.shape[:2])
    r, g, b = to_planar(src)
    return from_planar(np.array([r, g, b, alpha]))

def extract_alpha(image: np.ndarray) -> np.ndarray:
    """Extract the alpha plane from an RGBA image.
    
    Note that this returns a copy, not a view. To manipulate the pixels
    in a <i>view</i> of the alpha plane, simply make a numpy slice, as
    in: <code>alpha_view = myimage[:,:,3]</code>.
    """
    assert len(image.shape) == 3 and image.shape[2] == 4
    return np.dsplit(image, 4)[3].copy()

def extract_rgb(image: np.ndarray) -> np.ndarray:
    """Extract the RGB planes from an RGBA image.
    
    Note that this returns a copy. If you wish to obtain a view that
    allows mutating pixels, simply use slicing instead. For
    example, to invert the colors of an image while leaving alpha
    intact, you can do:
    <code>myimage[:,:,:3] = 1.0 - myimage[:,:,:3]</code>.
    """
    assert len(image.shape) == 3 and image.shape[2] >= 3
    planes = np.dsplit(image, image.shape[2])
    return np.dstack(planes[:3])

def to_planar(image: np.ndarray) -> np.ndarray:
    """Convert a row-major image into a channel-major image.
    
    This creates a copy, not a view.
    """
    assert len(image.shape) == 3
    result = np.array(np.dsplit(image, image.shape[2]))
    return np.reshape(result, result.shape[:-1])

def from_planar(image: np.ndarray) -> np.ndarray:
    """Create a channel-major image into row-major image.
    
    This creates a copy, not a view.
    """
    assert len(image.shape) == 3
    return np.dstack(image)



"""Tiny module for manipulating and generating floating-point images.

The Snowy API is intentionally flat and only defines functions, no
classes. Users should simply import the top-level package, not any of
its submodules.
"""

from .filtering import *
from .io import *
from .ops import *
from .distance import *
from .noise import *

__all__ = '''
GAUSSIAN HERMITE LANCZOS MITCHELL NEAREST TRIANGLE
blur resize

LINEAR SRGB GAMMA
delinearize
ensure_alpha
extract_alpha
extract_rgb
from_planar
linearize
load
reshape
export
show
to_planar
unshape
ColorSpace

add_border
compose
compose_premultiplied
gradient
hflip
hstack
rotate
unitize
vflip
vstack

generate_gdf
generate_sdf
generate_udf

generate_noise
'''.split()

# deprecated functions:
save = export



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="snowy",
    version="0.0.4",
    author="Philip Rideout",
    description="Small Image Library for Python 3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.6',
    install_requires=[
        'imageio>=2.3',
        'numpy>=1.14',
        'numba>=0.39',
    ],
    url="https://github.com/prideout/snowy",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)



import numpy as np
import math
from . import io

def generate_noise(width, height, frequency, seed=1, wrapx=False,
                   wrapy=False, offset=[0,0]):
    """Generate a single-channel gradient noise image.
    
    A frequency of 1.0 creates a single surflet across the width of the
    image, while a frequency of 4.0 creates a 4x4 grid such that the
    (2,2) surflet is centered. Noise values live within the [-1,+1]
    range.
    """
    return _noise(width, height, frequency, seed, wrapx, wrapy, offset)

def _noise(width, height, frequency, seed, wrapx, wrapy, offset):
    nrows, ncols = int(height), int(width)
    table = Noise(seed)

    # Compute the span of U texcoords in [0,+1] such that 0 is at the
    # left edge of the left-most texel, and +1 is at the right edge of
    # the right-most pixel.
    maxx = frequency
    hw = 0.5 * maxx / width
    u = np.linspace(hw, maxx - hw, ncols) + offset[0]

    # Compute the span of V texcoords according to the aspect ratio.
    maxy = frequency * float(height) / width
    hh = 0.5 * maxy / height
    v = np.linspace(hh, maxy - hh, nrows) + offset[1]

    # Generate floating point texture coordinates, then split them into
    # integer and fractional components.
    u, v = np.meshgrid(u, v, sparse=True)
    i0, j0 = np.floor(u).astype(int), np.floor(v).astype(int)
    i1, j1 = i0 + 1, j0 + 1
    x0, y0 = u - i0, v - j0
    x1, y1 = x0 - 1, y0 - 1

    # Find the 2D vectors at the nearest grid cell corners.
    if wrapx:
        assert math.modf(frequency)[0] == 0.0, \
            "wrapx requires an integer frequency"
        i0 = i0 % int(frequency)
        i1 = i1 % int(frequency)
    if wrapy:
        assert math.modf(maxy)[0] == 0.0, \
            "wrapy requires frequency*width/height to be an integer"
        j0 = j0 % int(maxy)
        j1 = j1 % int(maxy)
    grad00 = _gradient(table, i0, j0)
    grad01 = _gradient(table, i0, j1)
    grad10 = _gradient(table, i1, j0)
    grad11 = _gradient(table, i1, j1)

    va = dot(x0, y0, grad00[0], grad00[1])
    vb = dot(x1, y0, grad10[0], grad10[1])
    vc = dot(x0, y1, grad01[0], grad01[1])
    vd = dot(x1, y1, grad11[0], grad11[1])

    # Lerp the neighboring 4 surflets
    t0 = x0*x0*x0*(x0*(x0*6.0 - 15.0) + 10.0)
    t1 = y0*y0*y0*(y0*(y0*6.0 - 15.0) + 10.0)
    result = va + t0 * (vb-va) + t1 * (vc-va) + t0 * t1 * (va-vb-vc+vd)
    return io.reshape(result)

class Noise:
    def __init__(self, seed):
        self.rnd = np.random.RandomState(seed)
        self.size = 256
        self.mask = int(self.size - 1)
        self.indices = np.arange(self.size, dtype = np.int16)
        self.rnd.shuffle(self.indices)
        theta = np.linspace(0, math.tau, self.size, endpoint=False)
        self.gradients = [np.cos(theta), np.sin(theta)]

def _gradient(table: Noise, i, j):
    perm, mask = table.indices, table.mask
    u, v = table.gradients
    hash = perm[np.bitwise_and(perm[np.bitwise_and(i, mask)] + j, mask)]
    return u[hash], v[hash]

def dot(x, y, gradx, grady):
    return gradx * x + grady * y



"""Define add_border etc."""

from snowy.io import *
from numba import jit, guvectorize
import numpy as np

def add_left(image: np.ndarray, T=2, V=0) -> np.ndarray:
    height, width, nchan = image.shape
    newshape = height, width + T, nchan
    result = np.full(newshape, np.float64(V))
    np.copyto(result[:,T:], image)
    return result

def add_right(image: np.ndarray, T=2, V=0) -> np.ndarray:
    height, width, nchan = image.shape
    newshape = height, width + T, nchan
    result = np.full(newshape, np.float64(V))
    np.copyto(result[:,:-T], image)
    return result

def add_top(image: np.ndarray, T=2, V=0) -> np.ndarray:
    height, width, nchan = image.shape
    newshape = height + T, width, nchan
    result = np.full(newshape, np.float64(V))
    np.copyto(result[T:,:], image)
    return result

def add_bottom(image: np.ndarray, T=2, V=0) -> np.ndarray:
    height, width, nchan = image.shape
    newshape = height + T, width, nchan
    result = np.full(newshape, np.float64(V))
    np.copyto(result[:-T,:], image)
    return result

def add_border(image: np.ndarray, width=2, value=0, sides='ltrb'):
    """Extend the size of an image by adding borders.

    <p>
    The <code>sides</code> argument defaults to
    <code>"LTRB"</code>, which enables borders for all four sides: Left,
    Top, Right, and Bottom. This can be used to select which borders you
    wish to add.
    </p>

    """
    result = image
    sides = sides.upper()
    if 'L' in sides: result = add_left(result, width, value)
    if 'T' in sides: result = add_top(result, width, value)
    if 'R' in sides: result = add_right(result, width, value)
    if 'B' in sides: result = add_bottom(result, width, value)
    return result

def hstack(images, border_width=2, border_value=0):
    """Horizontally concatenate a list of images with a border.
    
    This is similar to numpy's <code>hstack</code> except that it adds
    a border around each image. The borders can be controlled
    with the optional <code>border_width</code> and
    <code>border_value</code> arguments. See also <a href="#vstack">
    vstack</a>.
    """
    if border_width == 0: return np.hstack(images)
    T, V = border_width, border_value
    result = []
    for image in images[:-1]:
        result.append(add_border(image, T, V, 'LTB'))
    result.append(add_border(images[-1], T, V))
    return np.hstack(result)

def vstack(images, border_width=2, border_value=0):
    """Vertically concatenate a list of images with a border.
    
    This is similar to numpy's <code>vstack</code> except that it adds
    a border around each image. The borders can be controlled
    with the optional <code>border_width</code> and
    <code>border_value</code> arguments. See also <a href="#hstack">
    hstack</a>.
    """
    if border_width == 0: return np.vstack(images)
    T, V = border_width, border_value
    result = []
    for image in images[:-1]:
        result.append(add_border(image, T, V, 'LTR'))
    result.append(add_border(images[-1], T, V))
    return np.vstack(result)

def unitize(img):
    """Remap the values so that they span the range from 0 to +1."""
    return (img - np.amin(img)) / (np.amax(img) - np.amin(img))

def gradient(img):
    """Compute X derivatives and Y derivatives."""
    nx, ny = np.gradient(unshape(img))
    return reshape(nx), reshape(ny)

def rotate(source: np.ndarray, degrees) -> np.ndarray:
    """Rotate image counter-clockwise by a multiple of 90 degrees."""
    assert len(source.shape) == 3, 'Shape is not rows x cols x channels'
    assert source.dtype == np.float, 'Images must be doubles.'
    h, w, c = source.shape
    degrees %= 360
    if degrees == 90:
        result = np.empty([w, h, c])
        rotate90(result, source)
    elif degrees == 180:
        result = np.empty([h, w, c])
        rotate180(result, source)
    elif degrees == 270:
        result = np.empty([w, h, c])
        rotate270(result, source)
    else:
        assert False, 'Angle must be a multiple of 90.'
    return result

def hflip(source: np.ndarray) -> np.ndarray:
    """Horizontally mirror the given image."""
    assert len(source.shape) == 3, 'Shape is not rows x cols x channels'
    assert source.dtype == np.float, 'Images must be doubles.'
    h, w, c = source.shape
    result = np.empty([h, w, c])
    jit_hflip(result, source)
    return result

def vflip(source: np.ndarray) -> np.ndarray:
    """Vertically mirror the given image."""
    assert len(source.shape) == 3, 'Shape is not rows x cols x channels'
    assert source.dtype == np.float, 'Images must be doubles.'
    h, w, c = source.shape
    result = np.empty([h, w, c])
    jit_vflip(result, source)
    return result

def compose(dst: np.ndarray, src: np.ndarray) -> np.ndarray:
    """Compose a source image with alpha onto a destination image."""
    a, b = ensure_alpha(src), ensure_alpha(dst)
    alpha = extract_alpha(a)
    result = b * (1.0 - alpha) + a * alpha
    if dst.shape[2] == 3:
        return extract_rgb(result)
    return result

def compose_premultiplied(dst: np.ndarray, src: np.ndarray):
    """Draw an image with premultiplied alpha over the destination."""
    a, b = ensure_alpha(src), ensure_alpha(dst)
    alpha = extract_alpha(a)
    result = b * (1.0 - alpha) + a
    if dst.shape[2] == 3:
        return extract_rgb(result)
    return result

SIG0 = "void(f8[:,:,:], f8[:,:,:])"
SIG1 = "(r,c,d),(c,r,d)"
@guvectorize([SIG0], SIG1, target='parallel')
def rotate90(result, source):
    nrows, ncols, nchan = source.shape
    for row in range(nrows):
        for col in range(ncols):
            for chan in range(nchan):
                v = source[row][col][chan]
                result[-col-1][row][chan] = v

SIG0 = "void(f8[:,:,:], f8[:,:,:])"
SIG1 = "(r,c,d),(r,c,d)"
@guvectorize([SIG0], SIG1, target='parallel')
def rotate180(result, source):
    nrows, ncols, nchan = source.shape
    for row in range(nrows):
        for col in range(ncols):
            for chan in range(nchan):
                v = source[row][col][chan]
                result[-row-1][-col-1][chan] = v

SIG0 = "void(f8[:,:,:], f8[:,:,:])"
SIG1 = "(r,c,d),(c,r,d)"
@guvectorize([SIG0], SIG1, target='parallel')
def rotate270(result, source):
    nrows, ncols, nchan = source.shape
    for row in range(nrows):
        for col in range(ncols):
            for chan in range(nchan):
                v = source[row][col][chan]
                result[col][-row-1][chan] = v

SIG0 = "void(f8[:,:,:], f8[:,:,:])"
SIG1 = "(r,c,d),(r,c,d)"
@guvectorize([SIG0], SIG1, target='parallel')
def jit_hflip(result, source):
    nrows, ncols, nchan = source.shape
    for row in range(nrows):
        for col in range(ncols):
            for chan in range(nchan):
                v = source[row][col][chan]
                result[row][-col-1][chan] = v

SIG0 = "void(f8[:,:,:], f8[:,:,:])"
SIG1 = "(r,c,d),(r,c,d)"
@guvectorize([SIG0], SIG1, target='parallel')
def jit_vflip(result, source):
    nrows, ncols, nchan = source.shape
    for row in range(nrows):
        for col in range(ncols):
            for chan in range(nchan):
                v = source[row][col][chan]
                result[-row-1][col][chan] = v
